from __future__ import annotations

from collections.abc import Callable, Sequence
from typing import Any

from asgiref.typing import ASGI3Application, HTTPScope, WebSocketScope
from asphalt.core import Context, current_context, resolve_reference
from starlette.applications import Starlette
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import Receive, Scope, Send

from asphalt.web.asgi3 import ASGIComponent


class AsphaltMiddleware(BaseHTTPMiddleware):
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        async with Context() as ctx:
            if scope["type"] == "http":
                ctx.add_resource(scope, types=[HTTPScope])
            elif scope["type"] == "websocket":
                ctx.add_resource(scope, types=[WebSocketScope])

            await super().__call__(scope, receive, send)

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        current_context().add_resource(request)
        return await call_next(request)


class StarletteComponent(ASGIComponent[Starlette]):
    """
    A component that serves a Starlette application.

    :param app: the Starlette application object
    :param host: the IP address to bind to
    :param port: the port to bind to
    :param debug: whether to enable debug mode in an implicitly created application
        (default: the value of
        `__debug__ <https://docs.python.org/3/library/constants.html#debug__>`_;
        ignored if an application object is explicitly passed in)
    :param middlewares: list of callables or dicts to be added as middleware using
        :meth:`add_middleware`
    """

    def __init__(
        self,
        components: dict[str, dict[str, Any] | None] = None,
        *,
        app: Starlette | None = None,
        host: str = "127.0.0.1",
        port: int = 8000,
        debug: bool | None = None,
        middlewares: Sequence[Callable[..., ASGI3Application] | dict[str, Any]] = (),
    ) -> None:
        debug = debug if isinstance(debug, bool) else __debug__
        super().__init__(
            components,
            app=app or Starlette(debug=debug),
            host=host,
            port=port,
            middlewares=middlewares,
        )

    def setup_asphalt_middleware(self, app: Starlette) -> ASGI3Application:
        return AsphaltMiddleware(app)

    def add_middleware(
        self, middleware: Callable[..., ASGI3Application] | dict[str, Any]
    ) -> None:
        """
        Add a middleware to the application.

        Unlike the raw ASGI version of this method, this one adds the middleware to
        Starlette's own middleware stack instead of wrapping the application directly.

        :param middleware: either a callable that takes the application object and
            returns an ASGI 3.0 application, or a dictionary containing a reference to
            such a callable. This dictionary must contain the key ``type`` which is a
            non-async callable (or a module:varname reference to one) and which will be
            called with the application object as the first positional argument and the
            rest of the keys in the dict as keyword arguments.

        """
        if isinstance(middleware, dict):
            type_ = resolve_reference(middleware.pop("type", None))
            if not callable(type_):
                raise TypeError(f"Middleware ({type_}) is not callable")

            self.app.add_middleware(type_, **middleware)
        elif callable(middleware):
            self.app.add_middleware(middleware)
        else:
            raise TypeError(
                f"middleware must be either a callable or a dict, not {middleware!r}"
            )



#!/usr/bin/env python3
from importlib.metadata import version

from packaging.version import parse

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx_autodoc_typehints",
    "sphinx_tabs.tabs",
]

templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
project = "asphalt-web"
author = "Alex Grönholm"
copyright = "2022, " + author

v = parse(version(project))
version = v.base_version
release = v.public

language = None

exclude_patterns = ["_build"]
pygments_style = "sphinx"
highlight_language = "python3"
todo_include_todos = False
autodoc_inherit_docstrings = False
autodoc_default_options = {"show-inheritance": True}

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
htmlhelp_basename = project.replace("-", "") + "doc"

extlinks = {
    "github": (
        f"https://github.com/asphalt-framework/{project}/tree/{release}/%s",
        None,
    )
}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "asphalt": ("https://asphalt.readthedocs.io/en/latest/", None),
}



from __future__ import annotations

from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from inspect import Signature, signature
from typing import Any, get_type_hints

from asgiref.typing import ASGI3Application
from asphalt.core import Context, require_resource, resolve_reference
from fastapi import Depends, FastAPI
from fastapi.routing import APIRoute, APIWebSocketRoute

from .asgi3 import ASGIComponent
from .starlette import AsphaltMiddleware


@dataclass
class _AsphaltDependency:
    name: str
    cls: type = field(init=False)

    async def __call__(self):
        return require_resource(self.cls, self.name)

    def __hash__(self) -> int:
        return hash((self.name, self.cls))

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, _AsphaltDependency):
            return self.name == other.name and self.cls == other.cls

        return NotImplemented


def AsphaltDepends(name: str = "default") -> Any:
    """
    Asphalt's version of FastAPI's :func:`~fastapi.Depends`.

    This should be marked as the default value on a parameter that should receive an
    Asphalt resource.

    :param name: the name of the resource within its unique type

    """
    return Depends(_AsphaltDependency(name))


class FastAPIComponent(ASGIComponent[FastAPI]):
    """
    A component that serves a FastAPI application.

    :param app: the FastAPI application object, or a module:varname reference to one
    :param host: the IP address to bind to
    :param port: the port to bind to
    :param debug: whether to enable debug mode in an implicitly created application
        (default: the value of
        `__debug__ <https://docs.python.org/3/library/constants.html#debug__>`_;
        ignored if an application object is explicitly passed in)
    :param middlewares: list of callables or dicts to be added as middleware using
        :meth:`add_middleware`
    """

    def __init__(
        self,
        components: dict[str, dict[str, Any] | None] = None,
        *,
        app: FastAPI | str | None = None,
        host: str = "127.0.0.1",
        port: int = 8000,
        debug: bool | None = None,
        middlewares: Sequence[Callable[..., ASGI3Application] | dict[str, Any]] = (),
    ) -> None:
        debug = debug if isinstance(debug, bool) else __debug__
        super().__init__(
            components,
            app=app or FastAPI(debug=debug),
            host=host,
            port=port,
            middlewares=middlewares,
        )

    def setup_asphalt_middleware(self, app: FastAPI) -> ASGI3Application:
        return AsphaltMiddleware(app)

    def add_middleware(
        self, middleware: Callable[..., ASGI3Application] | dict[str, Any]
    ) -> None:
        """
        Add a middleware to the application.

        Unlike the raw ASGI version of this method, this one adds the middleware to
        FastAPI's own middleware stack instead of wrapping the application directly.

        :param middleware: either a callable that takes the application object and
            returns an ASGI 3.0 application, or a dictionary containing a reference to
            such a callable. This dictionary must contain the key ``type`` which is a
            non-async callable (or a module:varname reference to one) and which will be
            called with the application object as the first positional argument and the
            rest of the keys in the dict as keyword arguments.

        """
        if isinstance(middleware, dict):
            type_ = resolve_reference(middleware.pop("type", None))
            if not callable(type_):
                raise TypeError(f"Middleware ({type_}) is not callable")

            self.app.add_middleware(type_, **middleware)
        elif callable(middleware):
            self.app.add_middleware(middleware)
        else:
            raise TypeError(
                f"middleware must be either a callable or a dict, not {middleware!r}"
            )

    async def start_server(self, ctx: Context) -> None:
        # Convert Asphalt dependencies into FastAPI dependencies
        for route in self.original_app.router.routes:
            if isinstance(route, (APIRoute, APIWebSocketRoute)):
                sig: Signature | None = None
                type_hints: dict[str, Any]
                for dependency in route.dependant.dependencies:
                    if isinstance(dependency.call, _AsphaltDependency):
                        if sig is None:
                            sig = signature(route.endpoint)
                            type_hints = get_type_hints(route.endpoint)

                        try:
                            annotation = type_hints[dependency.name]
                        except KeyError:
                            raise TypeError(
                                f"Dependency {dependency.name!r} in endpoint "
                                f"{route.path} is missing a type annotation"
                            ) from None

                        dependency.call.cls = annotation

        await super().start_server(ctx)



from __future__ import annotations

import pytest
from asgiref.typing import ASGI3Application
from asphalt.core import Context
from httpx import AsyncClient

try:
    from django.core.handlers.asgi import ASGIHandler

    from asphalt.web.django import DjangoComponent

    from .django_app.asgi import application
except ModuleNotFoundError:
    pytestmark = pytest.mark.skip("Django not available")


@pytest.mark.asyncio
async def test_http(unused_tcp_port: int):
    async with Context() as ctx, AsyncClient() as http:
        ctx.add_resource("foo")
        ctx.add_resource("bar", name="another")
        await DjangoComponent(app=application, port=unused_tcp_port).start(ctx)

        # Ensure that the application got added as a resource
        asgi_app = ctx.require_resource(ASGI3Application)
        asgi_handler = ctx.require_resource(ASGIHandler)
        assert asgi_handler is asgi_app

        response = await http.get(
            f"http://127.0.0.1:{unused_tcp_port}", params={"param": "Hello World"}
        )
        response.raise_for_status()
        assert response.json() == {
            "message": "Hello World",
            "my resource": "foo",
            "another resource": "bar",
        }



from __future__ import annotations

from collections.abc import AsyncIterator, Awaitable, Callable, Coroutine, Sequence
from inspect import iscoroutinefunction
from typing import Any

from aiohttp.web_app import Application
from aiohttp.web_middlewares import middleware
from aiohttp.web_request import Request
from aiohttp.web_response import Response
from aiohttp.web_runner import AppRunner, TCPSite
from asphalt.core import (
    ContainerComponent,
    Context,
    context_teardown,
    resolve_reference,
)


@middleware
async def asphalt_middleware(
    request: Request, handler: Callable[..., Awaitable]
) -> Response:
    async with Context() as ctx:
        ctx.add_resource(request, types=[Request])
        return await handler(request)


class AIOHTTPComponent(ContainerComponent):
    """
    A component that serves an aiohttp application.

    :param app: the application object, or a module:varname reference to one
    :param host: the IP address to bind to
    :param port: the port to bind to
    :param middlewares: list of compatible coroutine functions or dicts to be added as
        middleware using :meth:`add_middleware`
    """

    def __init__(
        self,
        components: dict[str, dict[str, Any] | None] = None,
        *,
        app: Application | str | None = None,
        host: str = "127.0.0.1",
        port: int = 8000,
        middlewares: Sequence[
            Callable[..., Coroutine[Any, Any, Any]] | dict[str, Any]
        ] = (),
    ) -> None:
        super().__init__(components)

        self.app = resolve_reference(app) or Application()
        self.host = host
        self.port = port

        self.add_middleware(asphalt_middleware)
        for mw in middlewares:
            self.add_middleware(mw)

    def add_middleware(
        self, middleware: Callable[..., Coroutine[Any, Any, Any]] | dict[str, Any]
    ) -> None:
        """
        Add middleware to the application.

        :param middleware: either a special coroutine function (as specified here_), or
            a dictionary containing a reference to a setup function. This dictionary
            must contain the key ``type`` which is a non-async callable (or a
            module:varname reference to one) and which will be called with the
            application object as the first positional argument and the rest of the keys
            in the dict as keyword arguments.

        .. _here: \
https://docs.aiohttp.org/en/stable/web_advanced.html#aiohttp-web-middlewares

        """
        if isinstance(middleware, dict):
            setup = resolve_reference(middleware.pop("type", None))
            if not callable(setup):
                raise TypeError(f"Setup function ({setup}) is not callable")

            setup(self.app, **middleware)
        elif iscoroutinefunction(middleware):
            self.app.middlewares.append(middleware)
        else:
            raise TypeError(
                f"middleware must be either a coroutine function or a dict, not "
                f"{middleware!r}"
            )

    async def start(self, ctx: Context) -> None:
        ctx.add_resource(self.app)
        await super().start(ctx)
        await self.start_server(ctx)

    @context_teardown
    async def start_server(self, ctx: Context) -> AsyncIterator[None]:
        """
        Start the HTTP server.

        This method is called by the component after the subcomponents have been
        started. If you need to add any middleware that requires resources provided by
        subcomponents, you can override this method and call the superclass
        implementation after the middleware has been added.

        """
        runner = AppRunner(self.app)
        await runner.setup()
        site = TCPSite(runner, host=self.host, port=self.port)
        await site.start()

        yield

        await runner.cleanup()



from fastapi import FastAPI
from fastapi.responses import PlainTextResponse

application = FastAPI()


@application.get("/", response_class=PlainTextResponse)
async def root() -> str:
    return "Hello, world!"



from asphalt.core import Component, Context, inject, resource
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import PlainTextResponse, Response


async def root(request: Request) -> Response:
    return PlainTextResponse("Hello, world!")


class WebRootComponent(Component):
    @inject
    async def start(self, ctx: Context, app: Starlette = resource()) -> None:
        app.add_route("/", root)



from asphalt.core import Component, Context, inject, resource
from fastapi import FastAPI
from fastapi.responses import PlainTextResponse


async def root() -> str:
    return "Hello, world!"


class WebRootComponent(Component):
    @inject
    async def start(self, ctx: Context, app: FastAPI = resource()) -> None:
        app.add_api_route("/", root, response_class=PlainTextResponse)



from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import PlainTextResponse, Response

application = Starlette()


@application.route("/")
async def root(request: Request) -> Response:
    return PlainTextResponse("Hello, world!")



from asphalt.core import inject, resource
from django.http import HttpRequest, HttpResponse, JsonResponse


@inject
async def index(
    request: HttpRequest,
    my_resource: str = resource(),
    another_resource: str = resource("another"),
) -> HttpResponse:
    return JsonResponse(
        {
            "message": request.GET["param"],
            "my resource": my_resource,
            "another resource": another_resource,
        }
    )



from __future__ import annotations

import json
from collections.abc import Callable, Sequence
from typing import Any, cast
from urllib.parse import parse_qs

import pytest
import websockets
from asgiref.typing import (
    ASGI3Application,
    ASGIReceiveCallable,
    ASGISendCallable,
    ASGISendEvent,
    HTTPScope,
    Scope,
    WebSocketScope,
)
from asphalt.core import Context, current_context, inject, resource
from httpx import AsyncClient

from asphalt.web.asgi3 import ASGIComponent


@inject
async def application(
    scope: HTTPScope,
    receive: ASGIReceiveCallable,
    send: ASGISendCallable,
    my_resource: str = resource(),
    another_resource: str = resource("another"),
):
    if scope["type"] == "http":
        current_context().require_resource(HTTPScope)
        query = parse_qs(cast(bytes, scope["query_string"]))
        await receive()

        body = json.dumps(
            {
                "message": query[b"param"][0].decode(),
                "my resource": my_resource,
                "another resource": another_resource,
            }
        ).encode()

        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [
                    (b"content-type", b"application/json"),
                    (b"content-length", b"%d" % len(body)),
                ],
            }
        )
        await send(
            {
                "type": "http.response.body",
                "body": body,
                "more_body": False,
            }
        )
    elif scope["type"] == "websocket":
        current_context().require_resource(WebSocketScope)
        await receive()  # Receive connection
        await send(  # Accept connection
            {
                "type": "websocket.accept",
            }
        )
        packet = await receive()
        await send(
            {
                "type": "websocket.send",
                "text": json.dumps(
                    {
                        "message": "Hello " + packet["text"],
                        "my resource": my_resource,
                        "another resource": another_resource,
                    }
                ),
            }
        )


class TextReplacerMiddleware:
    def __init__(self, app: ASGI3Application, text: str, replacement: str):
        self.app = app
        self.text = text.encode()
        self.replacement = replacement.encode()

    async def __call__(
        self, scope: Scope, receive: ASGIReceiveCallable, send: ASGISendCallable
    ) -> None:
        start_event: ASGISendEvent

        async def wrapped_send(event: ASGISendEvent) -> None:
            nonlocal start_event

            if event["type"] == "http.response.start":
                start_event = event
                return
            elif event["type"] == "http.response.body":
                event["body"] = event["body"].replace(self.text, self.replacement)
                for i, (key, value) in enumerate(start_event["headers"]):
                    if key == b"content-length":
                        start_event["headers"][i] = key, b"%d" % len(event["body"])
                        break

                await send(start_event)

            await send(event)

        await self.app(scope, receive, wrapped_send)


@pytest.mark.asyncio
async def test_http(unused_tcp_port: int):
    async with Context() as ctx, AsyncClient() as http:
        ctx.add_resource("foo")
        ctx.add_resource("bar", name="another")
        await ASGIComponent(app=application, port=unused_tcp_port).start(ctx)

        # Ensure that the application got added as a resource
        ctx.require_resource(ASGI3Application)

        # Ensure that the application responds correctly to an HTTP request
        response = await http.get(
            f"http://127.0.0.1:{unused_tcp_port}", params={"param": "Hello World"}
        )
        response.raise_for_status()
        assert response.json() == {
            "message": "Hello World",
            "my resource": "foo",
            "another resource": "bar",
        }


@pytest.mark.asyncio
async def test_ws(unused_tcp_port: int):
    async with Context() as ctx:
        ctx.add_resource("foo")
        ctx.add_resource("bar", name="another")
        await ASGIComponent(app=application, port=unused_tcp_port).start(ctx)

        # Ensure that the application got added as a resource
        ctx.require_resource(ASGI3Application)

        # Ensure that the application works correctly with a websocket connection
        async with websockets.connect(f"ws://localhost:{unused_tcp_port}") as ws:
            await ws.send("World")
            response = json.loads(await ws.recv())
            assert response == {
                "message": "Hello World",
                "my resource": "foo",
                "another resource": "bar",
            }


@pytest.mark.parametrize("method", ["direct", "dict"])
@pytest.mark.asyncio
async def test_middleware(unused_tcp_port: int, method: str):
    middlewares: Sequence[Callable[..., ASGI3Application] | dict[str, Any]]
    if method == "direct":
        middlewares = [lambda app: TextReplacerMiddleware(app, "World", "Middleware")]
    else:
        middlewares = [
            {
                "type": f"{__name__}:TextReplacerMiddleware",
                "text": "World",
                "replacement": "Middleware",
            }
        ]

    async with Context() as ctx, AsyncClient() as http:
        ctx.add_resource("foo")
        ctx.add_resource("bar", name="another")
        await ASGIComponent(
            app=application, port=unused_tcp_port, middlewares=middlewares
        ).start(ctx)

        # Ensure that the application got added as a resource
        ctx.require_resource(ASGI3Application)

        # Ensure that the application responds correctly to an HTTP request
        response = await http.get(
            f"http://127.0.0.1:{unused_tcp_port}", params={"param": "Hello World"}
        )
        response.raise_for_status()
        assert response.json() == {
            "message": "Hello Middleware",
            "my resource": "foo",
            "another resource": "bar",
        }


def test_bad_middleware_type():
    with pytest.raises(
        TypeError,
        match="middleware must be either a callable or a dict, not 'foo'",
    ):
        ASGIComponent(app=application, middlewares=["foo"])


def test_bad_middleware_dict():
    with pytest.raises(TypeError, match=r"Middleware \(1\) is not callable"):
        ASGIComponent(app=application, middlewares=[{"type": 1}])



async def application(scope, receive, send):
    """Trivial example of a raw ASGI 3.0 application without a framework."""
    if scope["type"] == "http":
        await receive()
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [
                    [b"content-type", b"text/plain"],
                ],
            }
        )
        await send(
            {
                "type": "http.response.body",
                "body": b"Hello, world!",
                "more_body": False,
            }
        )



from __future__ import annotations

from asyncio import create_task, sleep
from collections.abc import AsyncIterator, Callable, Sequence
from dataclasses import dataclass
from inspect import isfunction
from typing import Any, Generic, TypeVar

import uvicorn
from asgiref.typing import (
    ASGI3Application,
    ASGIReceiveCallable,
    ASGISendCallable,
    HTTPScope,
    Scope,
    WebSocketScope,
)
from asphalt.core import (
    ContainerComponent,
    Context,
    context_teardown,
    resolve_reference,
)
from uvicorn import Config

T_Application = TypeVar("T_Application", bound=ASGI3Application)


@dataclass
class AsphaltMiddleware:
    """
    Generic ASGI middleware for Asphalt integration.

    This middleware wraps both HTTP requests and websocket connections in their own
    contexts and exposes the ASGI scope object as a resource.

    :param asgiref.typing.ASGI3Application app: an ASGI 3.0 application
    """

    app: ASGI3Application

    async def __call__(
        self, scope: Scope, receive: ASGIReceiveCallable, send: ASGISendCallable
    ) -> None:
        if scope["type"] in ("http", "websocket"):
            async with Context() as ctx:
                scope_type = HTTPScope if scope["type"] == "http" else WebSocketScope
                ctx.add_resource(scope, types=[scope_type])
                await self.app(scope, receive, send)
        else:
            await self.app(scope, receive, send)


class ASGIComponent(ContainerComponent, Generic[T_Application]):
    """
    A component that serves the given ASGI 3.0 application via Uvicorn.

    :param app: the ASGI application to serve, or a module:varname reference to one
    :type app: asgiref.typing.ASGI3Application | None
    :param host: the IP address to bind to
    :param port: the port to bind to
    :param middlewares: list of callables or dicts to be added as middleware using
        :meth:`add_middleware`
    """

    def __init__(
        self,
        components: dict[str, dict[str, Any] | None] = None,
        *,
        app: T_Application | str,
        host: str = "127.0.0.1",
        port: int = 8000,
        middlewares: Sequence[Callable[..., ASGI3Application] | dict[str, Any]] = (),
    ) -> None:
        super().__init__(components)
        self.app: T_Application = resolve_reference(app)
        self.original_app = self.app
        self.host = host
        self.port = port

        self.add_middleware(self.setup_asphalt_middleware)
        for middleware in middlewares:
            self.add_middleware(middleware)

    def setup_asphalt_middleware(self, app: T_Application) -> ASGI3Application:
        return AsphaltMiddleware(app)

    def add_middleware(
        self, middleware: Callable[..., ASGI3Application] | dict[str, Any]
    ) -> None:
        """
        Add middleware to the application.

        :param middleware: either a callable that takes the application object and
            returns an ASGI 3.0 application, or a dictionary containing a reference to
            such a callable. This dictionary must contain the key ``type`` which is a
            non-async callable (or a module:varname reference to one) and which will be
            called with the application object as the first positional argument and the
            rest of the keys in the dict as keyword arguments.

        """
        if isinstance(middleware, dict):
            type_ = resolve_reference(middleware.pop("type", None))
            if not callable(type_):
                raise TypeError(f"Middleware ({type_!r}) is not callable")

            self.app = type_(self.app, **middleware)
        elif callable(middleware):
            self.app = middleware(self.app)
        else:
            raise TypeError(
                f"middleware must be either a callable or a dict, not {middleware!r}"
            )

    async def start(self, ctx: Context) -> None:
        types = [ASGI3Application]
        if not isfunction(self.original_app):
            types.append(type(self.original_app))

        ctx.add_resource(self.original_app, types=types)
        await super().start(ctx)
        await self.start_server(ctx)

    @context_teardown
    async def start_server(self, ctx: Context) -> AsyncIterator[None]:
        """
        Start the HTTP server.

        This method is called by the component after the subcomponents have been
        started. If you need to add any middleware that requires resources provided by
        subcomponents, you can override this method and call the superclass
        implementation after the middleware has been added.

        """
        config = Config(
            app=self.app,
            host=self.host,
            port=self.port,
            use_colors=False,
            log_config=None,
            lifespan="off",
        )
        server = uvicorn.Server(config)
        server.install_signal_handlers = lambda: None
        server_task = create_task(server.serve())
        while not server.started:
            await sleep(0)

        yield

        server.should_exit = True
        await server_task



from aiohttp.abc import Request
from aiohttp.web_app import Application
from aiohttp.web_response import Response
from asphalt.core import Component, Context, inject, resource


async def root(request: Request) -> Response:
    return Response(text="Hello, world!")


class WebRootComponent(Component):
    @inject
    async def start(self, ctx: Context, app: Application = resource()) -> None:
        app.router.add_route("GET", "/", root)



from __future__ import annotations

import json
from collections.abc import Callable, Sequence
from typing import Any

import pytest
import websockets
from asgiref.typing import ASGI3Application, HTTPScope, WebSocketScope
from asphalt.core import Component, Context, inject, require_resource, resource
from fastapi import FastAPI
from httpx import AsyncClient
from starlette.requests import Request
from starlette.responses import JSONResponse, PlainTextResponse, Response
from starlette.websockets import WebSocket

from asphalt.web.fastapi import AsphaltDepends, FastAPIComponent

from .test_asgi3 import TextReplacerMiddleware


@pytest.mark.parametrize("method", ["static", "dynamic"])
@pytest.mark.asyncio
async def test_http(unused_tcp_port: int, method: str):
    async def root(
        request: Request,
        my_resource: str = AsphaltDepends(),
        another_resource: str = AsphaltDepends("another"),
    ) -> Response:
        require_resource(HTTPScope)
        require_resource(Request)
        return JSONResponse(
            {
                "message": request.query_params["param"],
                "my resource": my_resource,
                "another resource": another_resource,
            }
        )

    application = FastAPI()
    if method == "static":
        application.add_api_route("/", root)
        components = {}
    else:

        class RouteComponent(Component):
            @inject
            async def start(self, ctx: Context, app: FastAPI = resource()) -> None:
                app.add_api_route("/", root)

        components = {"myroutes": {"type": RouteComponent}}

    async with Context() as ctx, AsyncClient() as http:
        ctx.add_resource("foo")
        ctx.add_resource("bar", name="another")
        await FastAPIComponent(
            components=components, app=application, port=unused_tcp_port
        ).start(ctx)

        # Ensure that the application got added as a resource
        asgi_app = ctx.require_resource(ASGI3Application)
        fastapi_app = ctx.require_resource(FastAPI)
        assert fastapi_app is asgi_app

        response = await http.get(
            f"http://127.0.0.1:{unused_tcp_port}", params={"param": "Hello World"}
        )
        response.raise_for_status()
        assert response.json() == {
            "message": "Hello World",
            "my resource": "foo",
            "another resource": "bar",
        }


@pytest.mark.parametrize("method", ["static", "dynamic"])
@pytest.mark.asyncio
async def test_ws(unused_tcp_port: int, method: str):
    async def ws_root(
        websocket: WebSocket,
        my_resource: str = AsphaltDepends(),
        another_resource: str = AsphaltDepends("another"),
    ):
        require_resource(WebSocketScope)
        await websocket.accept()
        message = await websocket.receive_text()
        await websocket.send_json(
            {
                "message": f"Hello {message}",
                "my resource": my_resource,
                "another resource": another_resource,
            }
        )

    application = FastAPI()
    if method == "static":
        application.add_api_websocket_route("/ws", ws_root)
        components = {}
    else:

        class RouteComponent(Component):
            @inject
            async def start(self, ctx: Context, app: FastAPI = resource()) -> None:
                app.add_api_websocket_route("/ws", ws_root)

        components = {"myroutes": {"type": RouteComponent}}

    async with Context() as ctx:
        ctx.add_resource("foo")
        ctx.add_resource("bar", name="another")
        await FastAPIComponent(
            components=components, app=application, port=unused_tcp_port
        ).start(ctx)

        # Ensure that the application got added as a resource
        asgi_app = ctx.require_resource(ASGI3Application)
        fastapi_app = ctx.require_resource(FastAPI)
        assert fastapi_app is asgi_app

        async with websockets.connect(f"ws://localhost:{unused_tcp_port}/ws") as ws:
            await ws.send("World")
            response = json.loads(await ws.recv())
            assert response == {
                "message": "Hello World",
                "my resource": "foo",
                "another resource": "bar",
            }


@pytest.mark.asyncio
async def test_missing_type_annotation():
    async def bad_root(request: Request, bad_resource=AsphaltDepends()) -> Response:
        return Response("never seen")

    application = FastAPI()
    application.add_api_route("/", bad_root)

    async with Context() as ctx:
        component = FastAPIComponent(app=application)
        with pytest.raises(
            TypeError,
            match="Dependency 'bad_resource' in endpoint / is missing a type annotation",
        ):
            await component.start(ctx)


@pytest.mark.parametrize("method", ["direct", "dict"])
@pytest.mark.asyncio
async def test_middleware(unused_tcp_port: int, method: str):
    middlewares: Sequence[Callable[..., ASGI3Application] | dict[str, Any]]
    if method == "direct":
        middlewares = [lambda app: TextReplacerMiddleware(app, "World", "Middleware")]
    else:
        middlewares = [
            {
                "type": f"{__name__}:TextReplacerMiddleware",
                "text": "World",
                "replacement": "Middleware",
            }
        ]

    async def root(request: Request) -> Response:
        return PlainTextResponse("Hello World")

    application = FastAPI()
    application.add_api_route("/", root)
    async with Context() as ctx, AsyncClient() as http:
        await FastAPIComponent(
            port=unused_tcp_port, app=application, middlewares=middlewares
        ).start(ctx)

        # Ensure that the application responds correctly to an HTTP request
        response = await http.get(
            f"http://127.0.0.1:{unused_tcp_port}", params={"param": "Hello World"}
        )
        response.raise_for_status()
        assert response.text == "Hello Middleware"


def test_bad_middleware_type():
    with pytest.raises(
        TypeError,
        match="middleware must be either a callable or a dict, not 'foo'",
    ):
        FastAPIComponent(middlewares=["foo"])


def test_bad_middleware_dict():
    with pytest.raises(TypeError, match=r"Middleware \(1\) is not callable"):
        FastAPIComponent(middlewares=[{"type": 1}])



from __future__ import annotations

from collections.abc import Awaitable, Callable

from asgiref.typing import ASGI3Application, HTTPScope
from asphalt.core import Context
from django.core.handlers.asgi import ASGIHandler, ASGIRequest
from django.http import HttpRequest, HttpResponse
from django.utils.decorators import async_only_middleware

from .asgi3 import ASGIComponent


@async_only_middleware
def AsphaltMiddleware(get_response: Callable[[HttpRequest], Awaitable[HttpResponse]]):
    async def middleware(request: HttpRequest) -> HttpResponse:
        async with Context() as ctx:
            ctx.add_resource(request)
            if isinstance(request, ASGIRequest):
                ctx.add_resource(request.scope, types=[HTTPScope])

            return await get_response(request)

    return middleware


class DjangoComponent(ASGIComponent[ASGIHandler]):
    """
    A component that serves a Django application.

    :param django.core.handlers.asgi.ASGIHandler app: the Django ASGI handler object
    :param host: the IP address to bind to
    :param port: the port to bind to
    :param middlewares: list of callables or dicts to be added as middleware using
        :meth:`add_middleware`
    """

    def setup_asphalt_middleware(self, app: ASGIHandler) -> ASGI3Application:
        # In Django, the middleware must be explicitly added to MIDDLEWARE in
        # settings.py
        return app



from aiohttp.abc import Request
from aiohttp.web_app import Application
from aiohttp.web_response import Response
from aiohttp.web_routedef import RouteTableDef

routes = RouteTableDef()


@routes.get("/")
async def root(request: Request) -> Response:
    return Response(text="Hello, world!")


application = Application()
application.add_routes(routes)



from __future__ import annotations

import json

import pytest
import websockets
from asphalt.core import Component, Context, inject, require_resource, resource
from httpx import AsyncClient

try:
    from aiohttp.abc import Request
    from aiohttp.web_app import Application
    from aiohttp.web_middlewares import middleware
    from aiohttp.web_response import Response, json_response
    from aiohttp.web_ws import WebSocketResponse

    from asphalt.web.aiohttp import AIOHTTPComponent
except ModuleNotFoundError:
    pytestmark = pytest.mark.skip("aiohttp not available")


def setup_text_replacer(app, *, text: str, replacement: str) -> None:
    @middleware
    async def text_replacer(request: Request, handler) -> None:
        response = await handler(request)
        response.text = response.text.replace(text, replacement)
        return response

    app.middlewares.append(text_replacer)


@pytest.mark.parametrize("method", ["static", "dynamic"])
@pytest.mark.asyncio
async def test_http(unused_tcp_port: int, method: str):
    @inject
    async def root(
        request,
        my_resource: str = resource(),
        another_resource: str = resource("another"),
    ):
        return json_response(
            {
                "message": request.query["param"],
                "my resource": my_resource,
                "another resource": another_resource,
            }
        )

    application = Application()
    if method == "static":
        application.router.add_route("GET", "/", root)
        components = {}
    else:

        class RouteComponent(Component):
            async def start(self, ctx: Context) -> None:
                app = require_resource(Application)
                app.router.add_route("GET", "/", root)

        components = {"myroutes": {"type": RouteComponent}}

    async with Context() as ctx, AsyncClient() as http:
        ctx.add_resource("foo")
        ctx.add_resource("bar", name="another")
        await AIOHTTPComponent(
            components=components, app=application, port=unused_tcp_port
        ).start(ctx)

        # Ensure that the application got added as a resource
        ctx.require_resource(Application)

        response = await http.get(
            f"http://127.0.0.1:{unused_tcp_port}", params={"param": "Hello World"}
        )
        response.raise_for_status()
        assert response.json() == {
            "message": "Hello World",
            "my resource": "foo",
            "another resource": "bar",
        }


@pytest.mark.parametrize("method", ["static", "dynamic"])
@pytest.mark.asyncio
async def test_ws(unused_tcp_port: int, method: str):
    @inject
    async def ws_root(
        request,
        my_resource: str = resource(),
        another_resource: str = resource("another"),
    ):
        ws = WebSocketResponse()
        await ws.prepare(request)
        message = await ws.receive_str()
        await ws.send_json(
            {
                "message": f"Hello {message}",
                "my resource": my_resource,
                "another resource": another_resource,
            }
        )

    application = Application()
    if method == "static":
        application.router.add_route("GET", "/", ws_root)
        components = {}
    else:

        class RouteComponent(Component):
            async def start(self, ctx: Context) -> None:
                app = require_resource(Application)
                app.router.add_route("GET", "/", ws_root)

        components = {"myroutes": {"type": RouteComponent}}

    async with Context() as ctx:
        ctx.add_resource("foo")
        ctx.add_resource("bar", name="another")
        await AIOHTTPComponent(
            components=components, app=application, port=unused_tcp_port
        ).start(ctx)

        # Ensure that the application got added as a resource
        ctx.require_resource(Application)

        async with websockets.connect(f"ws://localhost:{unused_tcp_port}") as ws:
            await ws.send("World")
            response = json.loads(await ws.recv())
            assert response == {
                "message": "Hello World",
                "my resource": "foo",
                "another resource": "bar",
            }


@pytest.mark.parametrize("method", ["direct", "dict"])
@pytest.mark.asyncio
async def test_middleware(unused_tcp_port: int, method: str):
    @middleware
    async def text_replacer(request: Request, handler) -> None:
        response = await handler(request)
        response.text = response.text.replace("Hello World", "Hello Middleware")
        return response

    async def root(request: Request) -> Response:
        return Response(text="Hello World")

    application = Application()
    application.router.add_route("GET", "/", root)

    if method == "direct":
        middlewares = [text_replacer]
    else:
        middlewares = [
            {
                "type": f"{__name__}:setup_text_replacer",
                "text": "Hello World",
                "replacement": "Hello Middleware",
            }
        ]

    async with Context() as ctx, AsyncClient() as http:
        ctx.add_resource("foo")
        ctx.add_resource("bar", name="another")
        await AIOHTTPComponent(
            app=application, port=unused_tcp_port, middlewares=middlewares
        ).start(ctx)

        # Ensure that the application got added as a resource
        ctx.require_resource(Application)

        response = await http.get(
            f"http://127.0.0.1:{unused_tcp_port}", params={"param": "Hello World"}
        )
        response.raise_for_status()
        assert response.text == "Hello Middleware"


def test_bad_middleware_type():
    with pytest.raises(
        TypeError,
        match="middleware must be either a coroutine function or a dict, not 'foo'",
    ):
        AIOHTTPComponent(middlewares=["foo"])


def test_bad_middleware_dict():
    with pytest.raises(TypeError, match=r"Setup function \(1\) is not callable"):
        AIOHTTPComponent(middlewares=[{"type": 1}])
